import glob

from git_stacktrace.tests import base
from git_stacktrace import parse_trace


class TestParsePythonStacktrace(base.TestCase):
    trace3_expected = [
        ("../common/utils/geo_utils.py", 68, "get_ip_geo", "return get_geo_db().record_by_addr(ip_address)"),
        (
            "/mnt/virtualenv_A/local/lib/python2.7/site-packages/pygeoip/__init__.py",
            563,
            "record_by_addr",
            "ipnum = util.ip2long(addr)",
        ),
        (
            "/mnt/virtualenv_A/local/lib/python2.7/site-packages/pygeoip/util.py",
            36,
            "ip2long",
            "return int(binascii.hexlify(socket.inet_pton(socket.AF_INET6, ip)), 16)",
        ),
    ]

    def get_trace(self, number=3):
        with open("git_stacktrace/tests/examples/python%d.trace" % number) as f:
            trace = parse_trace.PythonTraceback(f.readlines())
        return trace

    def test_extract_traceback_from_file(self):
        # extract_python_traceback_from_file will raise an exception if it incorrectly parses a file
        for filename in glob.glob("git_stacktrace/tests/examples/python*.trace"):
            with open(filename) as f:
                traceback = parse_trace.PythonTraceback(f.readlines())
                if filename == "git_stacktrace/tests/examples/python3.trace":
                    self.assertEqual(self.trace3_expected, traceback.traceback_format())

    def test_str(self):
        with open("git_stacktrace/tests/examples/python3.trace") as f:
            expected = f.read()
        trace = self.get_trace()
        self.assertEqual(expected, str(trace))
        with open("git_stacktrace/tests/examples/python5.trace") as f:
            expected = f.read()
        trace = self.get_trace(number=5)
        self.assertEqual(expected, str(trace))
        with open("git_stacktrace/tests/examples/python6.trace") as f:
            expected = f.read()
        trace = self.get_trace(number=6)
        self.assertEqual(expected, str(trace))

    def test_exception(self):
        self.assertRaises(parse_trace.ParseException, parse_trace.PythonTraceback, "NOT A TRACEBACK")
        with open("git_stacktrace/tests/examples/java1.trace") as f:
            self.assertRaises(parse_trace.ParseException, parse_trace.PythonTraceback, f.readlines())

    def test_file_match(self):
        trace = self.get_trace()
        self.assertTrue(trace.file_match(trace.lines[0].trace_filename, ["common/utils/geo_utils.py"]))
        self.assertFalse(trace.file_match(trace.lines[0].trace_filename, ["common/utils/fake.py"]))


class TestParseJavaStacktrace(base.TestCase):
    def get_trace(self):
        with open("git_stacktrace/tests/examples/java1.trace") as f:
            trace = parse_trace.JavaTraceback(f.readlines())
        return trace

    def test_str(self):
        with open("git_stacktrace/tests/examples/java1.trace") as f:
            expected = f.read()
        trace = self.get_trace()
        self.assertEqual(expected, str(trace))

    def test_extract_traceback(self):
        trace = self.get_trace()
        line = trace.lines[1]
        self.assertEqual(106, line.line_number)
        self.assertEqual("java/io/FileInputStream.java", line.trace_filename)
        self.assertEqual("FileInputStream", line.class_name)
        line = trace.lines[2]
        self.assertEqual(15, line.line_number)
        self.assertEqual("com/devdaily/tests/ExceptionTest.java", line.trace_filename)
        self.assertEqual("ExceptionTest", line.class_name)

    def test_exception(self):
        self.assertRaises(parse_trace.ParseException, parse_trace.JavaTraceback, "NOT A TRACEBACK")
        for filename in glob.glob("git_stacktrace/tests/examples/python*.trace"):
            with open(filename) as f:
                self.assertRaises(parse_trace.ParseException, parse_trace.JavaTraceback, f.readlines())

    def test_file_match(self):
        trace = self.get_trace()
        self.assertTrue(
            trace.file_match(trace.lines[2].trace_filename, ["src/main/java/com/devdaily/tests/ExceptionTest.java"])
        )
        self.assertFalse(
            trace.file_match(trace.lines[2].trace_filename, ["src/main/java/com/devdaily/tests/Fake.java"])
        )


class TestLine(base.TestCase):
    def test_line(self):
        line_data = ("./file", 1, "foo", "pass")
        line = parse_trace.Line(*line_data)
        line.git_filename = "file"
        self.assertEqual(line_data, line.traceback_format())


class TestParseTrace(base.TestCase):
    def test_parse_trace(self):
        for filename in glob.glob("git_stacktrace/tests/examples/*.trace"):
            with open(filename) as f:
                try:
                    parse_trace.parse_trace(f.readlines())
                except Exception:
                    self.fail("Failed to parse '%s'" % filename)



import mock
import testtools

from git_stacktrace.tests import base
from git_stacktrace import git
from git_stacktrace import result

fake_commit_info = git.CommitInfo(
    summary="summary", subject="subject", body="body", url="url", author="author", date=None
)


class TestResult(base.TestCase):
    commit_hash = "hash1"

    def test_init_result(self):
        commit1 = result.Result(self.commit_hash)
        self.assertEqual(self.commit_hash, commit1.commit)
        with testtools.ExpectedException(AttributeError):
            commit1.commit = "test"

    def test_files(self):
        commit1 = result.Result(self.commit_hash)
        expected_files_modified = set(["file1:10", "file1"])
        expected_files_added = set(["file2", "file3"])
        expected_files_deleted = set(["file4"])
        commit1.add_file(git.GitFile("file1", "M"), line_number=10)
        commit1.add_file(git.GitFile("file1", "M"))
        commit1.add_file(git.GitFile("file2", "A"))
        commit1.add_file(git.GitFile("file3", "C"))
        commit1.add_file(git.GitFile("file4", "D"))
        self.assertEqual(expected_files_modified, commit1.files_modified)
        self.assertEqual(expected_files_added, commit1.files_added)
        self.assertEqual(expected_files_deleted, commit1.files_deleted)

    def test_lines(self):
        commit1 = result.Result(self.commit_hash)
        expected_lines_added = set(["pass", "1+1"])
        expected_lines_removed = set(["1+2"])
        commit1.lines_added.add("pass")
        commit1.lines_added.add("pass")
        commit1.lines_added.add("1+1")
        self.assertEqual(expected_lines_added, commit1.lines_added)
        commit1.lines_removed.add("1+2")
        commit1.lines_removed.add("1+2")
        self.assertEqual(expected_lines_removed, commit1.lines_removed)

    @mock.patch("git_stacktrace.git.get_commit_info")
    def test_str(self, mocked_git_info):
        mocked_git_info.return_value = fake_commit_info
        commit1 = result.Result(self.commit_hash)
        commit1.add_file(git.GitFile("file1", "M"))
        expected = "summary\nLink:        url\nFiles Modified:\n    - file1\n"
        commit1.add_file(git.GitFile("file2", "A"))
        expected = "summary\nLink:        url\nFiles Added:\n    - file2\nFiles Modified:\n    - file1\n"
        self.assertEqual(expected, str(commit1))
        commit1.lines_added.add("pass")
        expected = (
            "summary\nLink:        url\nFiles Added:\n    - file2\nFiles Modified:\n    - "
            'file1\nLines Added:\n    - "pass"\n'
        )
        commit1.lines_removed.add("True")
        expected = (
            "summary\nLink:        url\nFiles Added:\n    - file2\nFiles Modified:\n    - "
            'file1\nLines Added:\n    - "pass"\nLines Removed:\n    - "True"\n'
        )
        commit1.add_file(git.GitFile("file3", "D"), line_number=11)
        expected = (
            "summary\nLink:        url\nFiles Added:\n    - file2\nFiles Modified:\n    - "
            "file1\nFiles Deleted:\n    - file3:11\nLines Added:\n    - "
            '"pass"\nLines Removed:\n    - "True"\n'
        )
        self.assertEqual(expected, str(commit1))

    @mock.patch("git_stacktrace.git.get_commit_info")
    def test_str_no_url(self, mocked_git_info):
        mocked_git_info.return_value = git.CommitInfo("summary", "subject", "body", None, "author", None)
        commit1 = result.Result(self.commit_hash)
        commit1.add_file(git.GitFile("file1", "M"))
        expected = "summary\nFiles Modified:\n    - file1\n"
        self.assertEqual(expected, str(commit1))
        commit1.lines_added.add("pass")
        expected = 'summary\nFiles Modified:\n    - file1\nLines Added:\n    - "pass"\n'
        commit1.lines_removed.add("True")
        expected = (
            'summary\nFiles Modified:\n    - file1\nLines Added:\n    - "pass"\nLines ' 'Removed:\n    - "True"\n'
        )
        self.assertEqual(expected, str(commit1))

    def test_rank(self):
        commit1 = result.Result(self.commit_hash)
        self.assertEqual(0, commit1.rank())
        commit1.add_file(git.GitFile("file1", "M"))
        self.assertEqual(1, commit1.rank())
        commit1.add_file(git.GitFile("file2", "A"))
        self.assertEqual(4, commit1.rank())
        commit1.lines_added.add("pass")
        self.assertEqual(7, commit1.rank())
        commit1.add_file(git.GitFile("file3", "M"), line_number=12)
        self.assertEqual(12, commit1.rank())

    @mock.patch("git_stacktrace.git.get_commit_info")
    def test_dict(self, mocked_git_info):
        mocked_git_info.return_value = fake_commit_info
        commit1 = result.Result(self.commit_hash)
        commit1.add_file(git.GitFile("file1", "M"))
        commit1.add_file(git.GitFile("file2", "A"), line_number=12)
        commit1.add_file(git.GitFile("file3", "D"))
        commit1.lines_removed.add("True")
        commit1.lines_added.add("pass")
        expected = {
            "commit": "hash1",
            "files_added": ["file2:12"],
            "files_modified": ["file1"],
            "files_deleted": ["file3"],
            "body": "body",
            "date": None,
            "author": "author",
            "subject": "subject",
            "lines_added": ["pass"],
            "lines_removed": ["True"],
            "summary": "summary",
            "url": "url",
        }
        self.assertEqual(expected, dict(commit1))

    @mock.patch("git_stacktrace.git.get_commit_info")
    def test_git_info(self, mocked_git_info):
        mocked_git_info.return_value = fake_commit_info
        commit1 = result.Result(self.commit_hash)
        self.assertEqual(commit1.summary, "summary")
        self.assertEqual(commit1.subject, "subject")
        self.assertEqual(commit1.body, "body")
        self.assertEqual(commit1.url, "url")
        self.assertEqual(commit1.author, "author")
        self.assertIsNone(commit1.date)
        self.assertEqual(mocked_git_info.call_count, 1)


class TestResults(base.TestCase):
    def test_results(self):
        results = result.Results()
        commit1 = results.get_result("hash1")
        commit1.add_file(git.GitFile("file1", "A"))
        commit2 = results.get_result("hash1")
        self.assertEqual(commit1, commit2)

    def test_sorted_results(self):
        results = result.Results()
        commit2 = results.get_result("hash2")
        commit1 = results.get_result("hash1")
        commit1.add_file(git.GitFile("file1", "M"))
        expected = [commit1, commit2]
        self.assertEqual(expected, results.get_sorted_results())

    def test_sorted_results_inverse(self):
        results = result.Results()
        commit1 = results.get_result("hash1")
        commit1.add_file(git.GitFile("file1", "M"))
        commit2 = results.get_result("hash2")
        expected = [commit1, commit2]
        self.assertEqual(expected, results.get_sorted_results())

    @mock.patch("git_stacktrace.git.get_commit_info")
    def test_get_sorted_results_by_dict(self, mocked_git_info):
        mocked_git_info.return_value = fake_commit_info
        results = result.Results()
        commit2 = results.get_result("hash2")
        commit1 = results.get_result("hash1")
        commit1.add_file(git.GitFile("file1", "M"))
        expected = [dict(commit1), dict(commit2)]
        self.assertEqual(expected, results.get_sorted_results_by_dict())



"""Extract important filenames, lines, functions and code from stacktrace

Currently only supports python stacktraces
"""
from __future__ import print_function

import abc
import logging
import re
import traceback

log = logging.getLogger(__name__)


class ParseException(Exception):
    pass


class Line(object):
    """Track data for each line in stacktrace"""

    def __init__(
        self, filename, line_number, function_name, code, class_name=None, native_method=False, unknown_source=False
    ):
        self.trace_filename = filename
        self.line_number = line_number
        self.function_name = function_name
        self.code = code
        self.class_name = class_name  # Java specific
        self.native_method = native_method  # Java specific
        self.unknown_source = unknown_source  # Java specific
        self.git_filename = None

    def traceback_format(self):
        return (self.trace_filename, self.line_number, self.function_name, self.code)


class Traceback(object, metaclass=abc.ABCMeta):
    def __init__(self, blob):
        self.header = ""
        self.footer = ""
        self.lines = None
        self.extract_traceback(self.prep_blob(blob))

    def prep_blob(self, blob):
        """Cleanup input."""
        # remove empty lines
        if type(blob) == list:
            blob = [line for line in blob if line.strip() != ""]
            if len(blob) == 1:
                blob = blob[0].replace("\\n", "\n").split("\n")
        # Split by line
        if isinstance(blob, str):
            lines = blob.split("\n")
        elif type(blob) == list:
            if len(blob) == 1:
                lines = blob[0].split("\n")
            else:
                lines = [line.rstrip() for line in blob]
        else:
            message = "Unknown input format"
            log.debug("%s - '%s", message, blob)
            raise ParseException(message)
        return lines

    @abc.abstractmethod
    def extract_traceback(self, lines):
        """Extract language specific traceback"""
        return

    @abc.abstractmethod
    def format_lines(self):
        """format extracted traceback in same way as traceback."""
        return

    def __str__(self):
        return self.header + self.format_lines() + self.footer

    @abc.abstractmethod
    def file_match(self, trace_filename, git_files):
        """How to match a trace_filename to git_files.

        Generally this varies depending on which is a substring of the other
        """
        return


class PythonTraceback(Traceback):
    """Parse Traceback string."""

    FILE_LINE_START = '  File "'

    def extract_traceback(self, lines):
        """Convert traceback string into a traceback.extract_tb format"""
        # filter out traceback lines
        self.header = lines[0] + "\n"
        if lines[-1] and not lines[-1].startswith(" "):
            self.footer = lines[-1] + "\n"
        lines = [line.rstrip() for line in lines if line.startswith("  ")]
        # extract
        extracted = []
        code_line = False
        for i, line in enumerate(lines):
            if code_line:
                code_line = False
                continue
            words = line.split(", ")
            if words[0].startswith(self.FILE_LINE_START):
                if not (words[0].startswith('  File "') and words[1].startswith("line ") and words[2].startswith("in")):
                    message = "Something went wrong parsing stacktrace input."
                    log.debug("%s - '%s'", message, line)
                    raise ParseException(message)
                f = words[0].split('"')[1].strip()
                line_number = int(words[1].split(" ")[1])
                function_name = " ".join(words[2].split(" ")[1:]).strip()
                if len(lines) == i + 1 or lines[i + 1].startswith(self.FILE_LINE_START):
                    # Not all lines have code in the traceback
                    code = None
                else:
                    code_line = True
                    code = str(lines[i + 1].strip())

                try:
                    extracted.append(Line(filename=f, line_number=line_number, function_name=function_name, code=code))
                except IndexError:
                    raise ParseException("Incorrectly extracted traceback information")
        self.lines = extracted
        # Sanity check
        new_lines = traceback.format_list(self.traceback_format())
        new_lines = "\n".join([line.rstrip() for line in new_lines])
        lines = "\n".join(lines)
        if lines != new_lines or not self.lines:
            message = "Incorrectly extracted traceback information"
            logging.debug("%s: original != parsed\noriginal:\n%s\nparsed:\n%s", message, lines, new_lines)
            raise ParseException(message)

    def traceback_format(self):
        return [line.traceback_format() for line in self.lines]

    def format_lines(self):
        lines = self.traceback_format()
        return "".join(traceback.format_list(lines))

    def file_match(self, trace_filename, git_files):
        # trace_filename is substring of git_filename
        return [f for f in git_files if trace_filename.endswith(f)]


class JavaTraceback(Traceback):
    def extract_traceback(self, lines):
        if not lines[0].startswith("\t"):
            self.header = lines[0] + "\n"
        lines = [line for line in lines if line.startswith("\t")]
        extracted = []
        for line in lines:
            extracted.append(self._extract_line(line))
        self.lines = extracted
        if not self.lines:
            raise ParseException("Failed to parse stacktrace")

    def _extract_line(self, line_string):
        if not line_string.startswith("\t"):
            raise ParseException("Missing tab at beginning of line")

        native_method = False
        unknown_source = False
        if line_string.endswith("(Native Method)"):
            # "at java.io.FileInputStream.open(Native Method)"
            native_method = True
        if line_string.endswith("(Unknown Source)"):
            # $Lambda$5/1034627183.run(Unknown Source)
            unknown_source = True

        # split on ' ', '(', ')', ':'
        tokens = re.split(r" |\(|\)|:", line_string.strip())

        if tokens[0] != "at" or len(tokens) != 5:
            raise ParseException("Invalid Java Exception")

        path = tokens[1].split(".")
        filename = "/".join(path[:-2] + [tokens[2]])
        function_name = path[-1]
        if not native_method and not unknown_source:
            line_number = int(tokens[3])
        else:
            line_number = None
        class_name = path[-2]
        return Line(filename, line_number, function_name, None, class_name, native_method, unknown_source)

    def _format_line(self, line):
        split = line.trace_filename.split("/")
        path = ".".join(split[:-1])
        filename = split[-1]
        if line.native_method:
            return "\tat %s.%s.%s(Native Method)\n" % (path, line.class_name, line.function_name)
        if line.unknown_source:
            return "\tat %s.%s.%s(Unknown Source)\n" % (path, line.class_name, line.function_name)
        return "\tat %s.%s.%s(%s:%d)\n" % (path, line.class_name, line.function_name, filename, line.line_number)

    def format_lines(self):
        return "".join(map(self._format_line, self.lines))

    def file_match(self, trace_filename, git_files):
        # git_filename is substring of trace_filename
        return [f for f in git_files if f.endswith(trace_filename)]


class JavaScriptTraceback(Traceback):
    # This class matches a stacktrace that looks similar to https://v8.dev/docs/stack-trace-api

    def extract_traceback(self, lines):
        if not lines[0].startswith("\t"):
            self.header = lines[0] + "\n"
        lines = [line for line in lines if line.startswith("\t")]
        extracted = []
        for line in lines:
            extracted.append(self._extract_line(line))
        self.lines = extracted
        if not self.lines:
            raise ParseException("Failed to parse stacktrace")

    def _extract_line(self, line_string):
        pattern = r"\tat\s(?P<symbol>[^\s(]*)?(?:\s*)?\(?(?P<path>[^\s\?]+)(?:\S*)?\:(?P<line>\d+):(?:\d+)\)?"
        result = re.match(pattern, line_string)

        if result:
            frame = result.groupdict()
        else:
            log.debug("Failed to parse frame %s", line_string)
            raise ParseException

        return Line(frame.get("path"), frame.get("line"), frame.get("symbol"), None)

    def traceback_format(self):
        return [line.traceback_format() for line in self.lines]

    def format_lines(self):
        lines = self.traceback_format()
        return "".join(traceback.format_list(lines))

    def file_match(self, trace_filename, git_files):
        return [f for f in git_files if trace_filename.endswith(f)]


def parse_trace(traceback_string):
    languages = [PythonTraceback, JavaTraceback, JavaScriptTraceback]
    for language in languages:
        try:
            return language(traceback_string)
        except ParseException:
            log.debug("Failed to parse as %s", language)
            # Try next language
            continue
    raise ParseException("Unable to parse traceback")



from __future__ import print_function

import collections
import datetime
import re
import subprocess
import sys
import shlex
import os

import whatthepatch

SHA1_REGEX = re.compile(r"\b[0-9a-f]{40}\b")

CommitInfo = collections.namedtuple("CommitInfo", ["summary", "subject", "body", "url", "author", "date"])


class GitFile(object):
    """Track filename and if file was added/removed or modified."""

    ADDED = "A"
    COPY_EDIT = "C"
    DELETED = "D"
    MODIFIED = "M"
    RENAME_EDIT = "R"
    TYPE = "T"
    UNMERGED = "U"
    UNKNOWN = "X"
    VALID = frozenset([ADDED, DELETED, MODIFIED, COPY_EDIT, RENAME_EDIT, TYPE, UNMERGED, UNKNOWN])

    def __init__(self, filename, state=None):
        self.filename = filename
        if state not in GitFile.VALID:
            raise Exception("Invalid git file state: %s" % state)
        self.state = state

    def __repr__(self):
        return self.filename

    def __eq__(self, other):
        if isinstance(other, str):
            other_filename = other
        else:
            other_filename = other.filename
        return self.filename == other_filename


def run_command_status(*argv, **kwargs):
    if len(argv) == 1:
        argv = shlex.split(str(argv[0]))
    stdin = kwargs.pop("stdin", None)
    newenv = os.environ.copy()
    newenv["LANG"] = "C"
    newenv["LANGUAGE"] = "C"
    newenv.update(kwargs)
    p = subprocess.Popen(
        argv, stdin=subprocess.PIPE if stdin else None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=newenv
    )
    (out, nothing) = p.communicate(stdin)
    out = out.decode("utf-8", "replace")
    return (p.returncode, out.strip())


def run_command(*argv, **kwargs):
    (rc, output) = run_command_status(*argv, **kwargs)
    if rc != 0:
        print(argv, rc, output)
        raise Exception("Something went wrong running the command %s %s" % (argv, kwargs))
    return output


def files_touched(git_range):
    """Run git log --pretty="%H" --raw  git_range.

    Generate a dictionary of files modified by the commits in range
    """
    cmd = "git", "log", "--pretty=%H", "--raw", git_range
    data = run_command(*cmd)
    commits = collections.defaultdict(list)
    commit = None
    for line in data.splitlines():
        if SHA1_REGEX.match(line):
            commit = line
        elif line.strip():
            split_line = line.split("\t")
            filename = split_line[-1]
            state = split_line[0].split(" ")[-1][0]
            commits[commit].append(GitFile(filename, state))
    return commits


def pickaxe(snippet, git_range, filename=None):
    """Run git log -S <snippet> <git_range> <filename>

    Use git pickaxe to 'Look for differences that change the number of occurrences of the
    specified string'

    If filename is passed in only look in that file

    Return list of commits that modified that snippet
    """
    cmd = "git", "log", "-b", "--pretty=%H", "-S", str(snippet), git_range
    if filename:
        cmd = cmd + ("--", filename,)
    commits = run_command(*cmd).splitlines()
    commits = [(commit, line_removed(snippet, commit)) for commit in commits]
    # Couldn't find a good way to POSIX regex escape the code and use regex
    # pickaxe to match full lines, so filter out partial results here.
    # Filter out results that aren't a full line
    commits = [commit for commit in commits if commit[1] is not None]
    return commits


def line_removed(target_line, commit):
    """Given a commit tell if target_line was added or removed.

    True if line was removed
    False if added
    None if target_line wasn't found at all (because not a full line etc.)
    """
    cmd = "git", "log", "-1", "--format=", "-p", str(commit)
    diff = run_command(*cmd)
    for diff in whatthepatch.parse_patch(diff):
        for line in diff.changes:
            if target_line in line[2]:
                if line[0] is None:
                    # Line added
                    return False
                elif line[1] is None:
                    # Line removed
                    return True
    # target_line matched part of a line instead of a full line
    return None


def line_match(commit, traceback_line):
    """Return true if line_number was added to filename in commit"""

    cmd = "git", "log", "-1", "--format=", "-p", str(commit)
    diff = run_command(*cmd)
    for diff in whatthepatch.parse_patch(diff):
        if diff.header.new_path == traceback_line.git_filename:
            for line in diff.changes:
                if line[0] is None:
                    if line[1] == traceback_line.line_number:
                        return True
    return False


def format_one_commit(commit):
    result = []
    info = get_commit_info(commit)
    result.append(info.summary)
    if info.url:
        result.append("Link:        " + info.url)
    return "\n".join(result)


def get_commit_info(commit, color=True):
    # Only use color if output is a terminal
    if sys.stdout.isatty() and color:
        cmd_prefix = "git", "log", "--color", "-1"
    else:
        cmd_prefix = "git", "log", "-1"
    git_format = (
        "--format=%C(auto,yellow)commit %H%C(auto,reset)%n" "Commit Date: %cD%nAuthor:      %aN <%aE>%nSubject:     %s"
    )
    summary = run_command(*(cmd_prefix + (git_format, commit)))

    # Find phabricator URL
    cmd = "git", "log", "-1", "--pretty=%b", commit
    body = run_command(*cmd)
    url = None
    for line in body.splitlines():
        if line.startswith("Differential Revision:"):
            url = line.split(" ")[2]

    cmd = "git", "log", "-1", "--format=%ct|%aN <%aE>", str(commit)
    date, author = run_command(*cmd).split("|", 1)
    date = datetime.datetime.fromtimestamp(int(date))

    cmd = "git", "log", "-1", "--format=%s", str(commit)
    subject = run_command(*cmd)

    return CommitInfo(summary=summary, subject=subject, body=body, url=url, author=author, date=date)


def valid_range(git_range):
    """Make sure there are commits in the range

    Returns True or False
    """
    cmd = "git", "log", "--oneline", git_range
    data = run_command(*cmd)
    lines = data.splitlines()
    return len(lines) > 0


def convert_since(since, branch=None):
    cmd = "git", "log", "--pretty=%H", "--since=%s" % since
    if branch:
        cmd = cmd + (branch,)
    data = run_command(*cmd)
    lines = data.splitlines()
    if len(lines) == 0:
        raise Exception("Didn't find any commits in 'since' range, try updating your git repo")
    return "%s..%s" % (lines[-1], lines[0])


def files(git_range):
    commit = git_range.split(".")[-1]
    cmd = "git", "ls-tree", "-r", "--name-only", commit
    data = run_command(*cmd)
    files = data.splitlines()
    return files



from __future__ import print_function

import json
import logging
import os

from html import escape
from git_stacktrace import api
from html.parser import HTMLParser
from urllib.parse import parse_qs
from string import Template
from datetime import date, datetime

log = logging.getLogger(__name__)
dir_path = os.path.dirname(os.path.realpath(__file__))


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


class Args(object):
    @staticmethod
    def from_json_body(body):
        return Args(json.loads(body))

    @staticmethod
    def from_qs(query_string):
        return Args(parse_qs(query_string.lstrip("?")))

    def __init__(self, params):
        self.params = params

    def _get_field(self, field, default=""):
        val = self.params.get(field, [default])
        val = val[0] if isinstance(val, list) else val
        return HTMLParser().unescape(val)

    @property
    def type(self):
        return self._get_field("option-type")

    @property
    def range(self):
        return self._get_field("range")

    @property
    def branch(self):
        return self._get_field("branch")

    @property
    def since(self):
        return self._get_field("since")

    @property
    def trace(self):
        return self._get_field("trace")

    @property
    def fast(self):
        return self._get_field("fast") == "on"

    def validate(self):
        if not self.type:
            return None

        if self.type == "by-date":
            if not self.since:
                return "Missing `since` value. Plese specify a date."
            self.git_range = api.convert_since(self.since, branch=self.branch)
            if not api.valid_range(self.git_range):
                return "Found no commits in '%s'" % self.git_range
        elif self.type == "by-range":
            self.git_range = self.range
            if not api.valid_range(self.git_range):
                return "Found no commits in '%s'" % self.git_range
        else:
            return "Invalid `type` value. Expected `by-date` or `by-range`."
        return None

    def get_results(self):
        if self.trace:
            traceback = api.parse_trace(self.trace)
            return api.lookup_stacktrace(traceback, self.git_range, fast=self.fast)
        else:
            return None


class ResultsOutput(object):
    def __init__(self, args):
        self.cwd = os.getcwd()
        self.args = args
        try:
            self.messages = args.validate()
            self.results = args.get_results()
        except Exception as e:
            self.messages = str(e)
            self.results = None

    def results_as_json(self):
        if self.results is None:
            return json.dumps({"errors": self.messages, "commits": [],}).encode()
        elif len(self.results.results) == 0:
            return json.dumps({"errors": "No matches found", "commits": [],}).encode()
        else:
            return json.dumps(
                {"errors": None, "commits": self.results.get_sorted_results_by_dict(),}, default=json_serial
            ).encode()

    def results_as_html(self):
        if self.results and self.results.results:
            sorted_results = self.results.get_sorted_results()
            return "\n<hr/>\n".join(
                ["<pre><code>" + escape(str(result)) + "</code></pre>" for result in sorted_results]
            )
        else:
            return "\n<hr/>\n<pre><code>No results found.</code></pre>\n"

    def messages_as_html(self):
        if self.messages is None:
            return ""
        with open(os.path.join(dir_path, "templates", "messages.html")) as f:
            return Template(f.read()).substitute(messages=escape(self.messages))

    def render_page(self):
        optionType = "by-date" if not self.args.type else self.args.type
        with open(os.path.join(dir_path, "templates", "page.html")) as f:
            return (
                Template(f.read())
                .substitute(
                    pwd=escape(self.cwd),
                    messages=self.messages_as_html(),
                    range=escape(self.args.range),
                    branch=escape(self.args.branch),
                    since=escape(self.args.since),
                    trace=escape(self.args.trace),
                    fast="checked" if self.args.fast else "",
                    optionType=escape(optionType),
                    isByDate="true" if optionType == "by-date" else "false",
                    isByRange="true" if optionType == "by-range" else "false",
                    byDateClass="active" if optionType == "by-date" else "",
                    byRangeClass="active" if optionType == "by-range" else "",
                    results=self.results_as_html(),
                )
                .encode("utf-8")
            )


class GitStacktraceApplication(object):
    def __init__(self, environ, start_response):
        self.environ = environ
        self.start_response = start_response
        self.path = environ["PATH_INFO"]

    def __iter__(self):
        method = self.environ["REQUEST_METHOD"]
        if method == "GET":
            yield self.do_GET() or b""
        elif method == "POST":
            yield self.do_POST() or b""
        elif method == "HEAD":
            self._set_headers()
            yield b""
        else:
            self._set_headers(500)
            yield b""

    def _set_headers(self, code=200, content_type="text/html"):
        codes = {
            200: "200 OK",
            404: "404 Not Found",
        }
        self.start_response(codes.get(code, "500 Internal Server Error"), [("Content-type", content_type)])

    def _request_body(self):
        content_length = int(self.environ["CONTENT_LENGTH"], 0)
        return self.environ["wsgi.input"].read(content_length)

    def do_GET(self):
        if self.path == "/favicon.ico":
            self._set_headers()
        elif self.path == "/":
            try:
                args = Args.from_qs(self.environ["QUERY_STRING"])
                out = ResultsOutput(args).render_page()
                self._set_headers()
                return out
            except Exception:
                log.exception("Unable to render trace page as html")
                self._set_headers(500)
        else:
            self._set_headers(404)

    def do_POST(self):
        if self.path == "/":
            try:
                args = Args.from_json_body(self._request_body())
                out = ResultsOutput(args).results_as_json()
                self._set_headers(200, "application/json")
                return out
            except Exception as e:
                log.exception("Unable to load trace results as json")
                self._set_headers(500, "application/json")
                return json.dumps({"error": str(e)}).encode()
        else:
            self._set_headers(404, "application/json")


application = GitStacktraceApplication



import mock

from git_stacktrace.tests import base
from git_stacktrace import api
from git_stacktrace import git


class TestApi(base.TestCase):
    @mock.patch("git_stacktrace.git.convert_since")
    def test_convert_since(self, mocked_command):
        expected = "HASH1..HASH2"
        mocked_command.return_value = expected
        self.assertEqual(expected, api.convert_since("1.day"))

    @mock.patch("git_stacktrace.git.valid_range")
    def test_valid_range(self, mocked_command):
        expected = True
        mocked_command.return_value = expected
        self.assertEqual(expected, api.valid_range("hash1..hash2"))

        expected = False
        mocked_command.return_value = expected
        self.assertEqual(expected, api.valid_range("hash1..hash2"))

    def get_traceback(self, java=False):
        if java:
            with open("git_stacktrace/tests/examples/java1.trace") as f:
                traceback = api.parse_trace(f.readlines())
        else:
            with open("git_stacktrace/tests/examples/python3.trace") as f:
                traceback = api.parse_trace(f.readlines())
        return traceback

    def setup_mocks(self, mock_files, mock_files_touched):
        mock_files_touched.return_value = {"hash2": [git.GitFile("common/utils/geo_utils.py", "M")]}
        mock_files.return_value = ["common/utils/geo_utils.py"]

    @mock.patch("git_stacktrace.git.pickaxe")
    @mock.patch("git_stacktrace.git.files_touched")
    @mock.patch("git_stacktrace.git.files")
    @mock.patch("git_stacktrace.git.line_match")
    def test_lookup_stacktrace_python(self, mock_line_match, mock_files, mock_files_touched, mock_pickaxe):
        mock_files_touched.return_value = True
        mock_line_match.return_value = False
        traceback = self.get_traceback()
        self.setup_mocks(mock_files, mock_files_touched)
        self.assertEqual(
            0,
            api.lookup_stacktrace(traceback, "hash1..hash3", fast=False).get_sorted_results()[0]._line_numbers_matched,
        )
        self.assertEqual(3, mock_pickaxe.call_count)

    @mock.patch("git_stacktrace.git.pickaxe")
    @mock.patch("git_stacktrace.git.files_touched")
    @mock.patch("git_stacktrace.git.files")
    @mock.patch("git_stacktrace.git.line_match")
    def test_lookup_stacktrace_java(self, mock_line_match, mock_files, mock_files_touched, mock_pickaxe):
        mock_files_touched.return_value = True
        mock_line_match.return_value = True
        traceback = self.get_traceback(java=True)
        mock_files.return_value = ["devdaily/src/main/java/com/devdaily/tests/ExceptionTest.java"]
        mock_files_touched.return_value = {
            "hash2": [git.GitFile("devdaily/src/main/java/com/devdaily/tests/ExceptionTest.java", "M")]
        }
        self.assertEqual(
            2,
            api.lookup_stacktrace(traceback, "hash1..hash3", fast=False).get_sorted_results()[0]._line_numbers_matched,
        )
        self.assertEqual(0, mock_pickaxe.call_count)

    @mock.patch("git_stacktrace.git.pickaxe")
    @mock.patch("git_stacktrace.git.files_touched")
    @mock.patch("git_stacktrace.git.files")
    @mock.patch("git_stacktrace.git.line_match")
    def test_lookup_stacktrace_fast(self, mock_line_match, mock_files, mock_files_touched, mock_pickaxe):
        mock_files_touched.return_value = True
        traceback = self.get_traceback()
        self.setup_mocks(mock_files, mock_files_touched)
        api.lookup_stacktrace(traceback, "hash1..hash3", fast=True)
        self.assertEqual(1, mock_pickaxe.call_count)

    @mock.patch("git_stacktrace.git.pickaxe")
    @mock.patch("git_stacktrace.git.files_touched")
    @mock.patch("git_stacktrace.git.files")
    @mock.patch("git_stacktrace.git.line_match")
    def test_lookup_stacktrace_line_match(self, mock_line_match, mock_files, mock_files_touched, mock_pickaxe):
        mock_files_touched.return_value = True
        mock_line_match.return_value = True
        traceback = self.get_traceback()
        self.setup_mocks(mock_files, mock_files_touched)
        self.assertEqual(
            1,
            api.lookup_stacktrace(traceback, "hash1..hash3", fast=False).get_sorted_results()[0]._line_numbers_matched,
        )
        self.assertEqual(3, mock_pickaxe.call_count)



import mock

from git_stacktrace.tests import base
from git_stacktrace import git
from git_stacktrace import parse_trace


class TestGitFile(base.TestCase):
    def test_git_file(self):
        git_file = git.GitFile("file1", "M")
        self.assertEqual(git_file.filename, "file1")
        self.assertEqual(git_file.state, git.GitFile.MODIFIED)
        git_file = git.GitFile("file2", "A")
        self.assertEqual(git_file.filename, "file2")
        self.assertEqual(git_file.state, git.GitFile.ADDED)

    def test_git_file_invalid(self):
        self.assertRaises(Exception, git.GitFile, "file1", "x")

    def test_git_file_cmp(self):
        git_file1 = git.GitFile("file1", "M")
        git_file = git.GitFile("file1", "A")
        self.assertEqual(git_file, git_file1)
        self.assertEqual(git_file, u"file1")
        self.assertEqual(git_file, "file1")


class TestGit(base.TestCase):
    @mock.patch("git_stacktrace.git.run_command")
    def test_convert_since_fail(self, mocked_command):
        mocked_command.return_value = ""
        self.assertRaises(Exception, git.convert_since, "1.day")

    @mock.patch("git_stacktrace.git.run_command")
    def test_convert_since(self, mocked_command):
        mocked_command.return_value = "\n".join(
            [
                "de75c8dd27af30daef012a9902af4c39c4728710",
                "04a13ace3a3e490a5e1a74aae740f45fee6562c3",
                "c0497475799306eebcfd014657150daa9af9c488",
                "f301b355050f64ebc2e83ebffc583713113aee9b",
                "32eba9e2c389c427c5b7b2288353eaf0903d52c0",
            ]
        )
        expected = "32eba9e2c389c427c5b7b2288353eaf0903d52c0..de75c8dd27af30daef012a9902af4c39c4728710"
        self.assertEqual(expected, git.convert_since("1.day"))

    @mock.patch("git_stacktrace.git.run_command")
    def test_files_touched(self, mocked_command):
        mocked_command.return_value = "\n".join(
            [
                "1ca8dd2b178ef8f308849bac2b0eaecaf91abc70",
                "",
                ":100644 100644 bcd1234... 0123456... M	file0",
                ":100644 100644 abcd123... 1234567... C68	file1	file2",
                ":100644 100644 abcd123... 1234567... R86	file1	file3",
                ":000000 100644 0000000... 1234567... A	file4 space/log",
                ":100644 100644 f9731ae1d4... 6dc2860... M       test/file"
                ":100644 000000 1234567... 0000000... D	file5",
            ]
        )
        expected = {"1ca8dd2b178ef8f308849bac2b0eaecaf91abc70": ["file0", "file2", "file3", "file4 space/log", "file5"]}
        self.assertEqual(expected, git.files_touched("A..B"))

    @mock.patch("git_stacktrace.git.run_command")
    def test_line_match(self, mocked_command):
        mocked_command.return_value = "\n".join(
            [
                "diff --git a/test_api.py b/test_api.py",
                "index 73e79d1..884b953 100644",
                "--- a/test_api.py",
                "+++ b/test_api.py",
                "@@ -35,7 +35,9 @@ class TestApi(base.TestCase):",
                "     @mock.patch('git_stacktrace.git.pickaxe')",
                "     @mock.patch('git_stacktrace.git.files_touched')",
                "     @mock.patch('git_stacktrace.git.files')",
                "-    def test_lookup_stacktrace(self, mock_files, mock_files_touched, mock_pickaxe):",
                "+    @mock.patch('git_stacktrace.git.line_match')",
                "+    def test_lookup_stacktrace(self, mock_line_match, mock_files, mock_files_touched, mock_pickaxe):",
                "+        mock_files_touched.return_value = True",
                "         traceback = self.get_traceback()",
                "         self.setup_mocks(mock_files, mock_files_touched)",
            ]
        )
        filename = "test_api.py"
        line = parse_trace.Line(filename, 38, None, None)
        line.git_filename = filename
        self.assertTrue(git.line_match("hash1", line))
        line.line_number = 5
        self.assertFalse(git.line_match("hash1", line))

    @mock.patch("git_stacktrace.git.run_command")
    def test_pickaxe(self, mocked_command):
        mocked_command.return_value = ""

        git.pickaxe("for f in sorted(self.files_added):", "hash1..hash2")
        expected = ("git", "log", "-b", "--pretty=%H", "-S", u"for f in sorted(self.files_added):", "hash1..hash2")
        mocked_command.assert_called_with(*expected)

        git.pickaxe("if 'a/b' in z", "hash1..hash2")
        expected = ("git", "log", "-b", "--pretty=%H", "-S", u"if 'a/b' in z", "hash1..hash2")
        mocked_command.assert_called_with(*expected)

        git.pickaxe("for f in sorted(self.files_added):", "hash1..hash2", "filename")
        expected = (
            "git",
            "log",
            "-b",
            "--pretty=%H",
            "-S",
            u"for f in sorted(self.files_added):",
            "hash1..hash2",
            "--",
            "filename",
        )
        mocked_command.assert_called_with(*expected)



import mock

from git_stacktrace.tests import base
from git_stacktrace.server import Args


class TestApi(base.TestCase):
    def test_default_args(self):
        args = Args({})
        self.assertEqual("", args.trace)
        self.assertEqual("", args.branch)
        self.assertFalse(args.fast)

    def test_args_from_json(self):
        json = '{"branch": "master", "fast": "on"}'
        args = Args.from_json_body(json)
        self.assertEqual("", args.trace)
        self.assertEqual("master", args.branch)
        self.assertTrue(args.fast)

    def test_args_from_json_as_array_vals(self):
        json = '{"branch": ["master"], "fast": ["on"]}'
        args = Args.from_json_body(json)
        self.assertEqual("master", args.branch)
        self.assertTrue(args.fast)

    def test_args_from_querystring(self):
        qs = "?branch=master&fast=on"
        args = Args.from_qs(qs)
        self.assertEqual("", args.trace)
        self.assertEqual("master", args.branch)
        self.assertTrue(args.fast)

    def test_args_from_querystring_without_qmark(self):
        qs = "branch=master&fast=on"
        args = Args.from_qs(qs)
        self.assertEqual("master", args.branch)

    def test_args_from_querystring_multiple_vals(self):
        qs = "?branch=master&branch=feature"
        args = Args.from_qs(qs)
        self.assertEqual("master", args.branch)

    def test_unescape_args(self):
        qs = "?branch=origin%2Fmaster&fast=on"
        args = Args.from_qs(qs)
        self.assertEqual("origin/master", args.branch)

    def test_args_ok_when_empty(self):
        args = Args({})
        self.assertIsNone(args.validate())

    def test_args_message_for_invalid_optionType(self):
        args = Args({"option-type": "foo"})
        self.assertIn("Invalid `type` value", args.validate())

    def test_args_byDate_requires_since(self):
        args = Args({"option-type": "by-date"})
        self.assertIn("Missing `since` value", args.validate())

    @mock.patch("git_stacktrace.server.api.convert_since")
    @mock.patch("git_stacktrace.server.api.valid_range")
    def test_args_byDate_errors_on_invalid_range(self, mock_valid_range, mock_convert_since):
        mock_valid_range.return_value = False
        args = Args({"option-type": "by-date", "since": "1.day"})
        self.assertIn("Found no commits in", args.validate())

    @mock.patch("git_stacktrace.server.api.convert_since")
    @mock.patch("git_stacktrace.server.api.valid_range")
    def test_args_byDate_returns_none_for_good_range(self, mock_valid_range, mock_convert_since):
        mock_valid_range.return_value = True
        args = Args({"option-type": "by-date", "since": "1.day"})
        self.assertIsNone(args.validate())

    @mock.patch("git_stacktrace.server.api.valid_range")
    def test_args_byRange_errors_on_invalid_range(self, mock_valid_range):
        mock_valid_range.return_value = False
        args = Args({"option-type": "by-range"})
        self.assertIn("Found no commits in", args.validate())

    @mock.patch("git_stacktrace.server.api.valid_range")
    def test_args_byRange_returns_none_for_good_range(self, mock_valid_range):
        mock_valid_range.return_value = True
        args = Args({"option-type": "by-range"})
        self.assertIsNone(args.validate())



from __future__ import print_function

import argparse
import logging
import os
import select
import sys

import git_stacktrace
from git_stacktrace import api
from git_stacktrace import server
from wsgiref.simple_server import make_server


def main():
    usage = "git stacktrace [<options>] [<RANGE>] < stacktrace from stdin"
    description = "Lookup commits related to a given stacktrace."
    parser = argparse.ArgumentParser(usage=usage, description=description)
    range_group = parser.add_mutually_exclusive_group()
    range_group.add_argument(
        "--since", metavar="<date1>", help="show commits " "more recent than a specific date (from git-log)"
    )
    range_group.add_argument("range", nargs="?", help="git commit range to use")
    range_group.add_argument(
        "--server", action="store_true", help="start a " "webserver to visually interact with git-stacktrace"
    )
    parser.add_argument("--port", default=os.environ.get("GIT_STACKTRACE_PORT", 8080), type=int, help="Server port")
    parser.add_argument(
        "-f",
        "--fast",
        action="store_true",
        help="Speed things up by not running " "pickaxe if the file for a line of code cannot be found",
    )
    parser.add_argument(
        "-b",
        "--branch",
        nargs="?",
        help="Git branch. If using --since, use this to "
        "specify which branch to run since on. Runs on current branch by default",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%s version %s" % (os.path.split(sys.argv[0])[-1], git_stacktrace.__version__),
    )
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug logging")
    args = parser.parse_args()

    logging.basicConfig(format="%(name)s:%(funcName)s:%(lineno)s: %(message)s")
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    if args.server:
        print("Starting httpd on port %s..." % args.port)
        httpd = make_server("", args.port, server.application)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            sys.exit(0)

    if args.since:
        git_range = api.convert_since(args.since, branch=args.branch)
        print("commit range: %s" % git_range, file=sys.stderr)
    else:
        if args.range is None:
            print("Error: Missing range and since, must use one\n")
            parser.print_help()
            sys.exit(1)
        git_range = args.range

    if not api.valid_range(git_range):
        print("Found no commits in '%s'" % git_range)
        sys.exit(1)

    if not select.select([sys.stdin], [], [], 0.0)[0]:
        raise Exception("No input found in stdin")
    blob = sys.stdin.readlines()
    traceback = api.parse_trace(blob)

    print(traceback)

    results = api.lookup_stacktrace(traceback, git_range, fast=args.fast)

    for r in results.get_sorted_results():
        print("")
        print(r)

    if len(results.get_sorted_results()) == 0:
        print("No matches found")


if __name__ == "__main__":
    main()



"""
python API to call git stacktrace.


Example usage::

    from git_stacktrace import api

    traceback = api.parse_trace(traceback_string)
    git_range = api.convert_since('1.day')
    results = api.lookup_stacktrace(traceback, git_range)
    for r in results.get_sorted_results():
        if "Smith" in r.author
            print("")
            print(r)
"""

import logging

from git_stacktrace import git
from git_stacktrace import result
from git_stacktrace import parse_trace

log = logging.getLogger(__name__)

# So we can call api.parse_trace
parse_trace = parse_trace.parse_trace


def _longest_filename(matches):
    """find longest match by number of '/'."""
    return max(matches, key=lambda filename: len(filename.split("/")))


def _lookup_files(commit_files, git_files, traceback, results):
    """Populate results and line.git_filename."""
    for line in traceback.lines:
        matches = traceback.file_match(line.trace_filename, git_files)
        if matches:
            git_file = _longest_filename(matches)
            for commit, file_list in commit_files.items():
                if git_file in file_list:
                    git_file = file_list[file_list.index(git_file)]
                    line.git_filename = git_file.filename
                    line_number = None
                    if git.line_match(commit, line):
                        line_number = line.line_number
                    results.get_result(commit).add_file(git_file, line_number)
            if line.git_filename is None:
                line.git_filename = _longest_filename(matches)


def convert_since(since, branch=None):
    """Convert the git since format into a git range

    :param since: git formatted since value such as '1,day'
    :type since: str
    :param branch: git branch, such as 'origin/master'
    :type branch: str
    :returns: commit range, such as 'ab9f71a..c04140'
    :rtype: str
    """
    return git.convert_since(since, branch=branch)


def valid_range(git_range):
    """Make sure there are commits in the range

    Generate a dictionary of files modified by the commits in range

    :param git_range: range of commits, such as 'ab9f71a..c04140'
    :type git_range: str
    :rtype: bool
    """
    return git.valid_range(git_range)


def lookup_stacktrace(traceback, git_range, fast=False):
    """Lookup to see what commits in git_range could have caused the stacktrace.

    Pass in a stacktrace object and returns a results object.

    :param traceback: Traceback object
    :type traceback: git_stacktrace.parse_trace.Traceback
    :param git_range: git commit range
    :type git_range: str
    :param fast: If True, don't run pickaxe if cannot find the file in git.
    :type fast: bool
    :returns: results
    :rtype: git_stacktrace.result.Results
    """
    results = result.Results()

    commit_files = git.files_touched(git_range)
    git_files = git.files(git_range)
    _lookup_files(commit_files, git_files, traceback, results)

    for line in traceback.lines:
        commits = []
        if line.code and not (line.git_filename is None and fast is True):
            try:
                commits = git.pickaxe(line.code, git_range, line.git_filename)
            except Exception:
                # If this fails, move on
                if log.isEnabledFor(logging.DEBUG):
                    log.exception("pickaxe failed")
                continue
        for commit, line_removed in commits:
            if line_removed is True:
                results.get_result(commit).lines_removed.add(line.code)
            if line_removed is False:
                results.get_result(commit).lines_added.add(line.code)
    return results



import fixtures
import testtools


class TestCase(testtools.TestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        stdout = self.useFixture(fixtures.StringStream("stdout")).stream
        self.useFixture(fixtures.MonkeyPatch("sys.stdout", stdout))
        stderr = self.useFixture(fixtures.StringStream("stderr")).stream
        self.useFixture(fixtures.MonkeyPatch("sys.stderr", stderr))



from git_stacktrace import git


class Result(object):
    """Track matches to stacktrace in a given commit."""

    def __init__(self, commit):
        self.__commit = commit
        self.files_modified = set()
        self.files_deleted = set()
        self.files_added = set()
        self.lines_added = set()
        self.lines_removed = set()
        self._line_numbers_matched = 0
        self.__commit_info_fetched = False

    def _lazy_fetch(self):
        if not self.__commit_info_fetched:
            self.__info = git.get_commit_info(self.commit, color=False)
            self.__commit_info_fetched = True

    @property
    def commit(self):
        """git commit hash"""
        return self.__commit

    @property
    def summary(self):
        """pre-formatted summary of git commit information"""
        self._lazy_fetch()
        return self.__info.summary

    @property
    def subject(self):
        """commit subject"""
        self._lazy_fetch()
        return self.__info.subject

    @property
    def body(self):
        """body of commit message"""
        self._lazy_fetch()
        return self.__info.body

    @property
    def url(self):
        """url found in commit body"""
        self._lazy_fetch()
        return self.__info.url

    @property
    def author(self):
        """commit author"""
        self._lazy_fetch()
        return self.__info.author

    @property
    def date(self):
        """commit date (not author date)"""
        self._lazy_fetch()
        return self.__info.date

    def add_file(self, git_file, line_number=None):
        if line_number:
            self._line_numbers_matched += 1
            filename = git_file.filename + ":" + str(line_number)
        else:
            filename = git_file.filename
        if git_file.state in [git.GitFile.ADDED, git.GitFile.COPY_EDIT]:
            self.files_added.add(filename)
        elif git_file.state in [git.GitFile.DELETED]:
            self.files_deleted.add(filename)
        else:
            self.files_modified.add(filename)

    def __hash__(self):
        return hash(self.commit)

    def __str__(self):
        result = ""
        result += git.format_one_commit(self.commit) + "\n"
        if len(self.files_added) > 0:
            result += "Files Added:\n"
        for f in sorted(self.files_added):
            result += "    - %s\n" % f

        if len(self.files_modified) > 0:
            result += "Files Modified:\n"
        for f in sorted(self.files_modified):
            result += "    - %s\n" % f

        if len(self.files_deleted) > 0:
            result += "Files Deleted:\n"
        for f in sorted(self.files_deleted):
            result += "    - %s\n" % f

        if len(self.lines_added) > 0:
            result += "Lines Added:\n"
        for line in sorted(self.lines_added):
            result += '    - "%s"\n' % line

        if len(self.lines_removed) > 0:
            result += "Lines Removed:\n"
        for line in sorted(self.lines_removed):
            result += '    - "%s"\n' % line
        return result

    def __iter__(self):
        self._lazy_fetch()
        yield "commit", self.commit
        yield "summary", self.summary
        yield "subject", self.subject
        yield "body", self.body
        yield "url", self.url
        yield "author", self.author
        yield "date", self.date
        yield "files_added", list(self.files_added)
        yield "files_modified", list(self.files_modified)
        yield "files_deleted", list(self.files_deleted)
        yield "lines_added", list(self.lines_added)
        yield "lines_removed", list(self.lines_removed)

    def rank(self):
        return (
            len(self.files_modified)
            + len(self.files_deleted) * 2
            + len(self.files_added) * 3
            + len(self.lines_added) * 3
            + len(self.lines_removed) * 2
            + self._line_numbers_matched * 4
        )

    def __eq__(self, other):
        return self.commit == other.commit

    def __lt__(self, other):
        if self.rank() == other.rank():
            # Make sorted order deterministic (but random) if rank is same
            # TODO fall back to sorting chronologically
            return self.commit < other.commit
        return self.rank() < other.rank()


class Results(object):
    """List of results."""

    def __init__(self):
        self.results = {}

    def get_result(self, commit):
        if commit not in self.results:
            self.results[commit] = Result(commit)
        return self.results[commit]

    def get_sorted_results(self):
        """Return list of results sorted by rank"""
        results = self.results.values()
        return sorted(results, reverse=True)

    def get_sorted_results_by_dict(self):
        """Return a list of dictionaries of the results sorted by rank"""
        results = self.get_sorted_results()
        return [dict(r) for r in results]
